/************************************************************************
	main.c

    RGB Luxeon Star Mood Light
    Copyright (C) 2009 Simon Inns

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	Email: simon.inns@gmail.com

************************************************************************/

// Global includes
#include <htc.h>

// PIC 18F1320 fuse configuration:
// Config word 1 (Oscillator configuration)
__CONFIG(1, 0xFEFF); // Compiler is broken (V9.63PL3) - HSPLL & FCMDIS & IESODIS
// Config word 2
__CONFIG(2, PWRTDIS & BORDIS & WDTDIS);
// Config word 3
__CONFIG(3, MCLRDIS);
// Config word 4
__CONFIG(4, STVRDIS & LVPDIS & DEBUGDIS);
// Config word 5, 6 and 7 (protection configuration)
__CONFIG(5, UNPROTECT);
__CONFIG(6, UNPROTECT);
__CONFIG(7, UNPROTECT);

// uController frequency (10Mhz * 4 PLL = 40Mhz)
#define _XTAL_FREQ 40000000

// Useful definitions
#define WRITE			0
#define READ			1
#define OFF				0
#define ON				1

// LED control pins
#define RED		RB0
#define GREEN	RA3
#define BLUE	RB1

// Max, Half and Quarter intensities
/*
#define MAX		63
#define HAF		16
#define QUA		8
*/
#define MAX		128
#define HAF		32
#define QUA		16

#define ONSPEED 600
#define OFFSPEED 600

// Define a structure to represent the lights
struct ledState {
	// Variables for foreground control
	int brightness;	// Target brightness
	int fadeOnSpeed;	// Fade ON speed 
	int fadeOffSpeed;	// Fade OFF speed 

	// Variables for interrupt control
	int currentBrightness;
	int fadeCounter;
} led[3];

// Define globals for interrupt handling
unsigned char globalPWMcounter = 0;

// High priority interrupt procedure
void interrupt hpHandler(void)
{
	// Nothing to do here at the moment
}

// Low priority interrupt procedure
void interrupt low_priority lpHandler(void)
{
	// Is this timer0 interrupting?
	if (TMR0IF)
	{
		// Increment the PWM counter
		globalPWMcounter++;
		if (globalPWMcounter == MAX) globalPWMcounter = 0;

		// Control the LED fading

		// fadeOnSpeed and fadeOffSpeed:
		// These control the speed at which the LED reaches the desired brightness (or dimness).
		// If the fadeSpeed is 0 the LED goes straight to full brightness
		// If the fadeSpeed is 1 the LED takes 1*64*208uS to get to full brightness (1/75th of a second)
		// If the fadeSpeed is 75 the LED takes approx. 1 second to get to full brightness, etc.

		unsigned char ledNumber;

		for (ledNumber = 0; ledNumber < 3; ledNumber++)
		{
			// Increment the fadeCounter
			led[ledNumber].fadeCounter++;

			// Fading on?
			if (led[ledNumber].currentBrightness < led[ledNumber].brightness)
			{
				if (led[ledNumber].fadeOnSpeed == 0)
				{
					led[ledNumber].currentBrightness = led[ledNumber].brightness;
					led[ledNumber].fadeCounter = 0;
				}
				else
				{
					if (led[ledNumber].fadeCounter >= led[ledNumber].fadeOnSpeed)
					{
						led[ledNumber].currentBrightness++;
						led[ledNumber].fadeCounter = 0;
					}
				}
			}

			// Fading off?
			if (led[ledNumber].currentBrightness >= led[ledNumber].brightness)
			{
				if (led[ledNumber].fadeOffSpeed == 0)
				{
					led[ledNumber].currentBrightness = led[ledNumber].brightness;
					led[ledNumber].fadeCounter = 0;
				}
				else
				{
					if (led[ledNumber].fadeCounter == led[ledNumber].fadeOffSpeed)
					{
						led[ledNumber].currentBrightness--;
						led[ledNumber].fadeCounter = 0;
					}
				}
			}

			// At the correct brightness?
			if (led[ledNumber].brightness == led[ledNumber].currentBrightness)
				led[ledNumber].fadeCounter = 0;
		}

		// Update the physical LED pins according to the PWM counter position
		if (led[0].currentBrightness > globalPWMcounter) RED = 1; else RED = 0;
		if (led[1].currentBrightness > globalPWMcounter) GREEN = 1; else GREEN = 0;
		if (led[2].currentBrightness > globalPWMcounter) BLUE = 1; else BLUE = 0;

		// Get ready for the next interrupt
		TMR0H = 0xFF;		// Reset the timer0 counter (High byte)
		TMR0L = 0x2F;		// Reset the timer0 counter (Low byte)
		TMR0IF = 0;			// Clear the timer0 interrupt flag
	}
}

// Millisecond delay (accurate to 10mS with a minimum of 10mS)
void delayMs(int milliseconds)
{
	// Timing is performed by timer1
	//
	// Fosc is 40Mhz or 40,000,000 ticks per second, input timing to the timer modules
	// is Fosc/4 or 10,000,000 ticks per second.  There are 1,000 mS per second
	// so 10,000 ticks = 1mS.  At a prescale of 1:2 there are 50,000 ticks per 10mS
	// so 65,535 - 50,000 = 15,535 or 0x3CAF)

	int tensOfMilliseconds;
	int tensLoop;
	
	// Divide the time down into tens of milliseconds and the remaining milliseconds
	tensOfMilliseconds = milliseconds / 10;
	milliseconds = milliseconds - (tensOfMilliseconds * 10);

	// Time the 10 millisecond blocks
	for (tensLoop = 0; tensLoop < tensOfMilliseconds; tensLoop++)
	{
		TMR1H = 0x3C;
		TMR1L = 0xAF;
		TMR1IF = 0; // Reset timer1's flag
		T1CON = 0b10100101;
		while(TMR1IF == 0); // Wait for timer 1 to flag @ 10mS
	}
}

// Colour mix (in random order)
const unsigned char colour[][3] = {
	MAX,   0,   0,	// Red primary
	QUA, QUA, MAX,
	HAF, MAX, MAX,
	MAX, MAX, QUA,
	HAF, HAF, MAX,
	MAX, QUA, HAF,
	MAX, HAF,   0,
	QUA,   0, MAX,
	QUA, MAX,   0,
	MAX,   0, MAX,
	MAX, QUA, MAX,
	MAX, HAF, HAF,
	  0, MAX,   0,
	  0, MAX, MAX,
	MAX,   0, HAF,
	  0,   0, MAX,
	HAF, QUA, MAX,
	HAF, MAX, HAF,
	MAX, HAF, MAX,
	  0, HAF, MAX,
	QUA, MAX, MAX,
	MAX, QUA, QUA,
	QUA, MAX, QUA,
	HAF,   0, MAX,
	  0, MAX, HAF,
	MAX, MAX, HAF,
	QUA, MAX, HAF,
	HAF, MAX,   0,
	  0, MAX, QUA,
	MAX, HAF, QUA,
	MAX, QUA,   0,
	MAX,   0, QUA,
	HAF, MAX, QUA,
	  0, QUA, MAX,
	QUA, HAF, MAX,
	MAX, MAX,   0,
};

void main(void)
{
	// Turn off analogue ports
	ADCON1 = 0x7F;

	// Configure ports as inputs (1) or outputs(0)
	TRISA = 0b00000000;
	TRISB = 0b00000000;

	// Clear all ports
	PORTA = 0b00000000;
	PORTB = 0b00000000;

	// Enable interrupts with priority
	IPEN = 1;

	// Set up timer0 to control the PWM
	//
	// Fosc is 40Mhz or 40,000,000 ticks per second, input timing to the timer modules
	// is Fosc/4 or 10,000,000 ticks per second.  There are 1,000,000 uS per second
	// so 10 ticks = 1uS. Therefore 208uS is 10*208 = 2080 ticks at 1:1 prescale value.
	// Minimum prescale is 1:2 so we need an interrupt every 2080/2 = 1040 ticks.
	//
	// If we set timer0 to be 16-bit the maximum counter value is (2^16)-1 = 65535.  Therefore we
	// need to set the initial timer value to 65535 - 1040 = 65327 or 0xFBEF in hexidecimal

	TMR0IP = 0;			// Set timer0 interrupt to low priority
	TMR0IF = 0;			// Clear the timer0 interrupt flag
	TMR0H = 0xFB;		// Reset the timer0 counter (High byte)
	TMR0L = 0xEF;		// Reset the timer0 counter (Low byte)
	T0CON = 0b10000000; // Timer0 on, 16-bit with a 1:2 prescale value
	TMR0IE = 1;			// Enable the timer0 interrupt

	// Before we enable interrupts we must initialise the led states
	int ledNumber;
	for (ledNumber = 0; ledNumber < 3; ledNumber++)
	{
		led[ledNumber].brightness = 0;
		led[ledNumber].fadeOnSpeed = 40;
		led[ledNumber].fadeOffSpeed = 40;
		led[ledNumber].currentBrightness = 0;
		led[ledNumber].fadeCounter = 0;
	}

	// Enable interrupts
	GIEH = 1;			// Global enable all high priority interrupts
	GIEL = 1;			// Global enable all low priority interrupts

	int step, loop;

	// Power up demonstration
	led[0].brightness = MAX;
	delayMs(1500);
	led[0].brightness = 0;
	led[1].brightness = MAX;
	delayMs(1500);
	led[1].brightness = 0;
	led[2].brightness = MAX;
	delayMs(1500);
	led[2].brightness = 0;
	delayMs(1000);

	// Cycle throught the available colour mixes
	for (step = 0; step < 35; step++)
	{
		led[0].brightness = colour[step][0];
		led[1].brightness = colour[step][1];
		led[2].brightness = colour[step][2];

		delayMs(500);
	}
	led[0].brightness = 0;
	led[1].brightness = 0;
	led[2].brightness = 0;
	delayMs(1000);

	// Initialise the LED state model
	led[0].fadeOnSpeed = ONSPEED;
	led[0].fadeOffSpeed = OFFSPEED;
	led[0].brightness = 0;
	led[1].fadeOnSpeed = ONSPEED;
	led[1].fadeOffSpeed = OFFSPEED;
	led[1].brightness = 0;
	led[2].fadeOnSpeed = ONSPEED;
	led[2].fadeOffSpeed = OFFSPEED;
	led[2].brightness = 0;

	while(1)
	{
		for (step = 0; step < 35; step++)
		{
			led[0].brightness = colour[step][0];
			led[1].brightness = colour[step][1];
			led[2].brightness = colour[step][2];
	
			// Wait 30 seconds
			delayMs(30000);
		}
	}	
}